/*********************************************************************
* Prueba de la multiplicacion de matrices con OpenMP llamando a blas (multithreaded)
* Para comprobar el comportamiento de usar dos niveles de paralelismo en arabi y ben
*********************************************************************/
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <omp.h>
#include <mkl.h>
#include "io.h"

extern void dgemm();

void mult(a,fa,ca,lda,b,fb,cb,ldb,c,fc,cc,ldc)
double *a;
int fa,ca,lda;
double *b;
int fb,cb,ldb;
double *c;
int fc,cc,ldc;
{
  double s,alpha=1.,beta=0.;
  char transa='n',transb='n';
  dgemm(&transa,&transb,&fa,&cb,&ca,&alpha,a,&lda,b,&ldb,&beta,c,&ldc);
}

void multomp(double *a, double *b, double *c, int n) {
  int i,t,il,nl;

  #pragma omp parallel private(i,t,nl,il)
  {
    t=omp_get_num_threads();
    i=omp_get_thread_num();
    if(i<n%t)
    {
      nl=n/t+1;
    }
    else
    {
      nl=n/t;
    }
    if(i<=n%t)
    {
      il=(n/t+1)*i;
    }
    else
    {
      il=(n/t*i+n%t);
    }
    mult(&a[il],nl,n,n,b,n,n,n,&c[il],nl,n,n);
  }
}

void escribir(double *m, int t) {
  int i, j;

  for (i = 0; i < t; i++) {
    for (j = 0; j < t; j++)
      printf("%.4lf ", m[i * t + j]);
    printf("\n");
  }
  printf("\n");
}

/*
c
c     mseconds - returns elapsed microseconds since Jan 1st, 1970.
c
*/
double mseconds() {
  struct timeval t;
  gettimeofday(&t, NULL);
  return t.tv_sec * 1000000. + t.tv_usec;
}

int main(int argc, char *argv[]) {
  int t, iN, fN, incN, threadsomp, threadsmkl;
  double start, fin, tiempo;
  double *a, *b, *c;
  int ith,fth,incth,totalnth,ompdynamic,mkldynamic;

// valores iniciales, final e incremento de tamano, y numero maximo de threads
  iN = atoi(argv[1]);
  fN = atoi(argv[2]);
  incN = atoi(argv[3]);
  ith = atoi(argv[4]);
  fth = atoi(argv[5]);
  incth=atoi(argv[6]);
  ompdynamic=atoi(argv[7]);
  mkldynamic=atoi(argv[8]);

#ifdef TIME
          printf("#tamano, threadsomp, threadsmkl, tiemposeg, omp dynamic %d, mkl dynamic %d\n",ompdynamic,mkldynamic);
#endif

  omp_set_nested(1); //permite paralelismo anidado, aunque no se si hace falta porque los threads de omp y mkl parece que van por separado
  omp_set_dynamic(ompdynamic);
  mkl_set_dynamic(mkldynamic); //con 0, deshabilita que mkl pueda decidir el numero de threads que considera hasta el maximo que se le da con mkl_set_num_threads
  for(t = iN; t <= fN; t += incN) 
  {
    for(totalnth=ith;totalnth<=fth;totalnth+=incth)
    {
      for(threadsomp=1;threadsomp<=totalnth;threadsomp++)
      {
        if((totalnth%threadsomp)==0)
        {
          threadsmkl=totalnth/threadsomp;
          omp_set_num_threads(threadsomp);
          mkl_set_num_threads(threadsmkl);
          a = (double *) malloc(sizeof(double) * t * t);
          b = (double *) malloc(sizeof(double) * t * t);
          c = (double *) malloc(sizeof(double) * t * t);   
    
 
          generar_matriz(a,t,t,t,-10.,10.);
          generar_matriz(b,t,t,t,-10.,10.);
  //      generar_matriz_ld_fija(a,t,t,t);
  //      generar_matriz_ld_fija(b,t,t,t);
#ifdef DEBUG
    escribir(a, t);
    escribir(b, t);
#endif
          start = mseconds();

          multomp(a, b , c, t);

          fin = mseconds();
          tiempo = (fin - start) / 1000000.;

          if (tiempo == 0.) {
            printf("No hay suficiente precision\n");
          } else {
#ifdef TIME
          printf(" %d %d %d %d,%06d\n",t,threadsomp,threadsmkl,(int) tiempo,(int) ((tiempo-(int) tiempo)*1000000.));
#endif
#ifdef DEBUG
      escribir(c, t);
#endif    
          }
          free(a);
          free(b);
          free(c);
        }
      }
    }
  }
}
